@@ -849,16 +849,24 @@ class SubqueryLoader(AbstractRelationshipLoader):
         # to look only for significant columns
         q = orig_query._clone().correlate(None)
 
-        # set a real "from" if not present, as this is more
-        # accurate than just going off of the column expression
-        if not q._from_obj and orig_entity.is_mapper and \
-                orig_entity.mapper.isa(leftmost_mapper):
-            q._set_select_from([orig_entity], False)
-        target_cols = q._adapt_col_list(leftmost_attr)
+        # set the query's "FROM" list explicitly to what the
+        # FROM list would be in any case, as we will be limiting
+        # the columns in the SELECT list which may no longer include
+        # all entities mentioned in things like WHERE, JOIN, etc.
+        if not q._from_obj:
+            q._set_select_from(
+                list(set([
+                    ent['entity'] for ent in orig_query.column_descriptions
+                    if ent['entity'] is not None
+                ])),
+                False
+            )
 
-        # select from the identity columns of the outer.  This will remove
+        # select from the identity columns of the outer (specifically, these
+        # are the 'local_cols' of the property).  This will remove
         # other columns from the query that might suggest the right entity
-        # which is why we try to _set_select_from above.
+        # which is why we do _set_select_from above.
+        target_cols = q._adapt_col_list(leftmost_attr)
         q._set_entities(target_cols)
 
         distinct_target_key = leftmost_relationship.distinct_target_key
