@@ -1,4 +1,4 @@
-# Copyright (C) 2005-2016 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2017 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -23,72 +23,59 @@ class Load(Generative, MapperOption):
     :class:`.Query` in order to affect how various mapped attributes are
     loaded.
 
-    .. versionadded:: 0.9.0 The :meth:`.Load` system is a new foundation for
-       the existing system of loader options, including options such as
-       :func:`.orm.joinedload`, :func:`.orm.defer`, and others.   In
-       particular, it introduces a new method-chained system that replaces the
-       need for dot-separated paths as well as "_all()" options such as
-       :func:`.orm.joinedload_all`.
+    The :class:`.Load` object is in most cases used implicitly behind the
+    scenes when one makes use of a query option like :func:`.joinedload`,
+    :func:`.defer`, or similar.   However, the :class:`.Load` object
+    can also be used directly, and in some cases can be useful.
 
-    A :class:`.Load` object can be used directly or indirectly.  To use one
-    directly, instantiate given the parent class.  This style of usage is
-    useful when dealing with a :class:`.Query` that has multiple entities,
-    or when producing a loader option that can be applied generically to
-    any style of query::
+    To use :class:`.Load` directly, instantiate it with the target mapped
+    class as the argument.   This style of usage is
+    useful when dealing with a :class:`.Query` that has multiple entities::
 
         myopt = Load(MyClass).joinedload("widgets")
 
-    The above ``myopt`` can now be used with :meth:`.Query.options`::
+    The above ``myopt`` can now be used with :meth:`.Query.options`, where it
+    will only take effect for the ``MyClass`` entity::
 
-        session.query(MyClass).options(myopt)
+        session.query(MyClass, MyOtherClass).options(myopt)
 
-    The :class:`.Load` construct is invoked indirectly whenever one makes use
-    of the various loader options that are present in ``sqlalchemy.orm``,
-    including options such as :func:`.orm.joinedload`, :func:`.orm.defer`,
-    :func:`.orm.subqueryload`, and all the rest.  These constructs produce an
-    "anonymous" form of the :class:`.Load` object which tracks attributes and
-    options, but is not linked to a parent class until it is associated with a
-    parent :class:`.Query`::
+    One case where :class:`.Load` is useful as public API is when specifying
+    "wildcard" options that only take effect for a certain class::
 
-        # produce "unbound" Load object
-        myopt = joinedload("widgets")
+        session.query(Order).options(Load(Order).lazyload('*'))
 
-        # when applied using options(), the option is "bound" to the
-        # class observed in the given query, e.g. MyClass
-        session.query(MyClass).options(myopt)
+    Above, all relationships on ``Order`` will be lazy-loaded, but other
+    attributes on those descendant objects will load using their normal
+    loader strategy.
 
-    Whether the direct or indirect style is used, the :class:`.Load` object
-    returned now represents a specific "path" along the entities of a
-    :class:`.Query`.  This path can be traversed using a standard
-    method-chaining approach.  Supposing a class hierarchy such as ``User``,
-    ``User.addresses -> Address``, ``User.orders -> Order`` and
-    ``Order.items -> Item``, we can specify a variety of loader options along
-    each element in the "path"::
-
-        session.query(User).options(
-                    joinedload("addresses"),
-                    subqueryload("orders").joinedload("items")
-                )
-
-    Where above, the ``addresses`` collection will be joined-loaded, the
-    ``orders`` collection will be subquery-loaded, and within that subquery
-    load the ``items`` collection will be joined-loaded.
+    .. seealso::
 
+        :ref:`loading_toplevel`
 
     """
 
     def __init__(self, entity):
         insp = inspect(entity)
         self.path = insp._path_registry
+        # note that this .context is shared among all descendant
+        # Load objects
         self.context = {}
         self.local_opts = {}
 
+    @classmethod
+    def for_existing_path(cls, path):
+        load = cls.__new__(cls)
+        load.path = path
+        load.context = {}
+        load.local_opts = {}
+        return load
+
     def _generate(self):
         cloned = super(Load, self)._generate()
         cloned.local_opts = {}
         return cloned
 
-    _merge_into_path = False
+    is_opts_only = False
     strategy = None
     propagate_to_loaders = False
 
@@ -201,7 +188,7 @@ class Load(Generative, MapperOption):
             self._set_path_strategy()
 
     @_generative
-    def set_column_strategy(self, attrs, strategy, opts=None):
+    def set_column_strategy(self, attrs, strategy, opts=None, opts_only=False):
         strategy = self._coerce_strat(strategy)
 
         for attr in attrs:
@@ -212,21 +199,34 @@ class Load(Generative, MapperOption):
             cloned.propagate_to_loaders = True
             if opts:
                 cloned.local_opts.update(opts)
+            if opts_only:
+                cloned.is_opts_only = True
             cloned._set_path_strategy()
 
-    def _set_path_strategy(self):
-        if self._merge_into_path:
-            # special helper for undefer_group
-            existing = self.path.get(self.context, "loader")
-            if existing:
-                existing.local_opts.update(self.local_opts)
-            else:
-                self.path.set(self.context, "loader", self)
+    def _set_for_path(self, context, path, replace=True, merge_opts=False):
+        if merge_opts or not replace:
+            existing = path.get(self.context, "loader")
 
-        elif self.path.has_entity:
-            self.path.parent.set(self.context, "loader", self)
+            if existing:
+                if merge_opts:
+                    existing.local_opts.update(self.local_opts)
+            else:
+                path.set(context, "loader", self)
         else:
-            self.path.set(self.context, "loader", self)
+            existing = path.get(self.context, "loader")
+            path.set(context, "loader", self)
+            if existing and existing.is_opts_only:
+                self.local_opts.update(existing.local_opts)
+
+    def _set_path_strategy(self):
+        if self.path.has_entity:
+            effective_path = self.path.parent
+        else:
+            effective_path = self.path
+
+        self._set_for_path(
+            self.context, effective_path, replace=True,
+            merge_opts=self.is_opts_only)
 
     def __getstate__(self):
         d = self.__dict__.copy()
@@ -314,7 +314,7 @@ class _UnboundLoad(Load):
             val._bind_loader(query, query._attributes, raiseerr)
 
     @classmethod
-    def _from_keys(self, meth, keys, chained, kw):
+    def _from_keys(cls, meth, keys, chained, kw):
         opt = _UnboundLoad()
 
         def _split_key(key):
@@ -399,6 +399,7 @@ class _UnboundLoad(Load):
         loader = Load(path_element)
         loader.context = context
         loader.strategy = self.strategy
+        loader.is_opts_only = self.is_opts_only
 
         path = loader.path
         for token in start_path:
@@ -420,24 +421,15 @@ class _UnboundLoad(Load):
 
         if effective_path.is_token:
             for path in effective_path.generate_for_superclasses():
-                if self._merge_into_path:
-                    # special helper for undefer_group
-                    existing = path.get(context, "loader")
-                    if existing:
-                        existing.local_opts.update(self.local_opts)
-                    else:
-                        path.set(context, "loader", loader)
-                elif self._is_chain_link:
-                    path.setdefault(context, "loader", loader)
-                else:
-                    path.set(context, "loader", loader)
+                loader._set_for_path(
+                    context, path,
+                    replace=not self._is_chain_link,
+                    merge_opts=self.is_opts_only)
         else:
-            # only supported for the undefer_group() wildcard opt
-            assert not self._merge_into_path
-            if self._is_chain_link:
-                effective_path.setdefault(context, "loader", loader)
-            else:
-                effective_path.set(context, "loader", loader)
+            loader._set_for_path(
+                context, effective_path,
+                replace=not self._is_chain_link,
+                merge_opts=self.is_opts_only)
 
     def _find_entity_prop_comparator(self, query, token, mapper, raiseerr):
         if _is_aliased_class(mapper):
@@ -537,7 +529,7 @@ See :func:`.orm.%(name)s` for usage examples.
 
 @loader_option()
 def contains_eager(loadopt, attr, alias=None):
-    """Indicate that the given attribute should be eagerly loaded from
+    r"""Indicate that the given attribute should be eagerly loaded from
     columns stated manually in the query.
 
     This function is part of the :class:`.Load` interface and supports
@@ -546,26 +538,40 @@ def contains_eager(loadopt, attr, alias=None):
     The option is used in conjunction with an explicit join that loads
     the desired rows, i.e.::
 
-        sess.query(Order).\\
-                join(Order.user).\\
+        sess.query(Order).\
+                join(Order.user).\
                 options(contains_eager(Order.user))
 
     The above query would join from the ``Order`` entity to its related
     ``User`` entity, and the returned ``Order`` objects would have the
     ``Order.user`` attribute pre-populated.
 
-    :func:`contains_eager` also accepts an `alias` argument, which is the
+    :func:`.contains_eager` also accepts an `alias` argument, which is the
     string name of an alias, an :func:`~sqlalchemy.sql.expression.alias`
     construct, or an :func:`~sqlalchemy.orm.aliased` construct. Use this when
     the eagerly-loaded rows are to come from an aliased table::
 
         user_alias = aliased(User)
-        sess.query(Order).\\
-                join((user_alias, Order.user)).\\
+        sess.query(Order).\
+                join((user_alias, Order.user)).\
                 options(contains_eager(Order.user, alias=user_alias))
 
+    When using :func:`.contains_eager` in conjunction with inherited
+    subclasses, the :meth:`.RelationshipProperty.of_type` modifier should
+    also be used in order to set up the pathing properly::
+
+        sess.query(Company).\
+            outerjoin(Company.employees.of_type(Manager)).\
+            options(
+                contains_eager(
+                    Company.employees.of_type(Manager),
+                    alias=Manager)
+            )
+
     .. seealso::
 
+        :ref:`loading_toplevel`
+
         :ref:`contains_eager`
 
     """
@@ -608,7 +614,7 @@ def load_only(loadopt, *attrs):
     ``Address`` object load only the ``email_address`` attribute::
 
         session.query(User).options(
-                subqueryload("addreses").load_only("email_address")
+                subqueryload("addresses").load_only("email_address")
         )
 
     For a :class:`.Query` that has multiple entities, the lead entity can be
@@ -652,11 +658,13 @@ def joinedload(loadopt, attr, innerjoin=None):
         query(User).options(joinedload(User.orders))
 
         # joined-load Order.items and then Item.keywords
-        query(Order).options(joinedload(Order.items).joinedload(Item.keywords))
+        query(Order).options(
+            joinedload(Order.items).joinedload(Item.keywords))
 
         # lazily load Order.items, but when Items are loaded,
         # joined-load the keywords collection
-        query(Order).options(lazyload(Order.items).joinedload(Item.keywords))
+        query(Order).options(
+            lazyload(Order.items).joinedload(Item.keywords))
 
     :param innerjoin: if ``True``, indicates that the joined eager load should
      use an inner join instead of the default of left outer join::
@@ -673,22 +681,22 @@ def joinedload(loadopt, attr, innerjoin=None):
 
      The above query, linking A.bs via "outer" join and B.cs via "inner" join
      would render the joins as "a LEFT OUTER JOIN (b JOIN c)".   When using
-     SQLite, this form of JOIN is translated to use full subqueries as this
-     syntax is otherwise not directly supported.
+     older versions of SQLite (< 3.7.16), this form of JOIN is translated to
+     use full subqueries as this syntax is otherwise not directly supported.
 
      The ``innerjoin`` flag can also be stated with the term ``"unnested"``.
-     This will prevent joins from being right-nested, and will instead
-     link an "innerjoin" eagerload to an "outerjoin" eagerload by bypassing
-     the "inner" join.   Using this form as follows::
+     This indicates that an INNER JOIN should be used, *unless* the join
+     is linked to a LEFT OUTER JOIN to the left, in which case it
+     will render as LEFT OUTER JOIN.  For example, supposing ``A.bs``
+     is an outerjoin::
 
         query(A).options(
-            joinedload(A.bs, innerjoin=False).
+            joinedload(A.bs).
                 joinedload(B.cs, innerjoin="unnested")
         )
 
-     Joins will be rendered as "a LEFT OUTER JOIN b LEFT OUTER JOIN c", so that
-     all of "a" is matched rather than being incorrectly limited by a "b" that
-     does not contain a "c".
+     The above join will render as "a LEFT OUTER JOIN b LEFT OUTER JOIN c",
+     rather than as "a LEFT OUTER JOIN (b JOIN c)".
 
      .. note:: The "unnested" flag does **not** affect the JOIN rendered
         from a many-to-many association table, e.g. a table configured
@@ -696,9 +704,6 @@ def joinedload(loadopt, attr, innerjoin=None):
         correctness of results, these joins are always INNER and are
         therefore right-nested if linked to an OUTER join.
 
-     .. versionadded:: 0.9.4 Added support for "nesting" of eager "inner"
-        joins.  See :ref:`feature_2976`.
-
      .. versionchanged:: 1.0.0 ``innerjoin=True`` now implies
         ``innerjoin="nested"``, whereas in 0.9 it implied
         ``innerjoin="unnested"``.  In order to achieve the pre-1.0 "unnested"
@@ -710,7 +715,8 @@ def joinedload(loadopt, attr, innerjoin=None):
         The joins produced by :func:`.orm.joinedload` are **anonymously
         aliased**.  The criteria by which the join proceeds cannot be
         modified, nor can the :class:`.Query` refer to these joins in any way,
-        including ordering.
+        including ordering.  See :ref:`zen_of_eager_loading` for further
+        detail.
 
         To produce a specific SQL JOIN which is explicitly available, use
         :meth:`.Query.join`.   To combine explicit JOINs with eager loading
@@ -721,16 +727,7 @@ def joinedload(loadopt, attr, innerjoin=None):
 
         :ref:`loading_toplevel`
 
-        :ref:`contains_eager`
-
-        :func:`.orm.subqueryload`
-
-        :func:`.orm.lazyload`
-
-        :paramref:`.relationship.lazy`
-
-        :paramref:`.relationship.innerjoin` - :func:`.relationship`-level
-        version of the :paramref:`.joinedload.innerjoin` option.
+        :ref:`joined_eager_loading`
 
     """
     loader = loadopt.set_relationship_strategy(attr, {"lazy": "joined"})
@@ -776,11 +773,7 @@ def subqueryload(loadopt, attr):
 
         :ref:`loading_toplevel`
 
-        :func:`.orm.joinedload`
-
-        :func:`.orm.lazyload`
-
-        :paramref:`.relationship.lazy`
+        :ref:`subquery_eager_loading`
 
     """
     return loadopt.set_relationship_strategy(attr, {"lazy": "subquery"})
@@ -806,7 +799,9 @@ def lazyload(loadopt, attr):
 
     .. seealso::
 
-        :paramref:`.relationship.lazy`
+        :ref:`loading_toplevel`
+
+        :ref:`lazy_loading`
 
     """
     return loadopt.set_relationship_strategy(attr, {"lazy": "select"})
@@ -834,12 +829,6 @@ def immediateload(loadopt, attr):
 
         :ref:`loading_toplevel`
 
-        :func:`.orm.joinedload`
-
-        :func:`.orm.lazyload`
-
-        :paramref:`.relationship.lazy`
-
     """
     loader = loadopt.set_relationship_strategy(attr, {"lazy": "immediate"})
     return loader
@@ -861,6 +850,10 @@ def noload(loadopt, attr):
     :func:`.orm.noload` applies to :func:`.relationship` attributes; for
     column-based attributes, see :func:`.orm.defer`.
 
+    .. seealso::
+
+        :ref:`loading_toplevel`
+
     """
 
     return loadopt.set_relationship_strategy(attr, {"lazy": "noload"})
@@ -872,22 +865,74 @@ def noload(*keys):
 
 
 @loader_option()
-def defaultload(loadopt, attr):
-    """Indicate an attribute should load using its default loader style.
+def raiseload(loadopt, attr, sql_only=False):
+    """Indicate that the given relationship attribute should disallow lazy loads.
 
-    This method is used to link to other loader options, such as
-    to set the :func:`.orm.defer` option on a class that is linked to
-    a relationship of the parent class being loaded, :func:`.orm.defaultload`
-    can be used to navigate this path without changing the loading style
-    of the relationship::
+    A relationship attribute configured with :func:`.orm.raiseload` will
+    raise an :exc:`~sqlalchemy.exc.InvalidRequestError` upon access.   The
+    typical way this is useful is when an application is attempting to ensure
+    that all relationship attributes that are accessed in a particular context
+    would have been already loaded via eager loading.  Instead of having
+    to read through SQL logs to ensure lazy loads aren't occurring, this
+    strategy will cause them to raise immediately.
 
-        session.query(MyClass).options(defaultload("someattr").defer("some_column"))
+    :param sql_only: if True, raise only if the lazy load would emit SQL,
+     but not if it is only checking the identity map, or determining that
+     the related value should just be None due to missing keys.  When False,
+     the strategy will raise for all varieties of lazyload.
+
+    This function is part of the :class:`.Load` interface and supports
+    both method-chained and standalone operation.
+
+    :func:`.orm.raiseload` applies to :func:`.relationship` attributes only.
+
+    .. versionadded:: 1.1
 
     .. seealso::
 
-        :func:`.orm.defer`
+        :ref:`loading_toplevel`
 
-        :func:`.orm.undefer`
+        :ref:`prevent_lazy_with_raiseload`
+
+    """
+
+    return loadopt.set_relationship_strategy(
+        attr, {"lazy": "raise_on_sql" if sql_only else "raise"})
+
+
+@raiseload._add_unbound_fn
+def raiseload(*keys, **kw):
+    return _UnboundLoad._from_keys(_UnboundLoad.raiseload, keys, False, kw)
+
+
+@loader_option()
+def defaultload(loadopt, attr):
+    """Indicate an attribute should load using its default loader style.
+
+    This method is used to link to other loader options further into
+    a chain of attributes without altering the loader style of the links
+    along the chain.  For example, to set joined eager loading for an
+    element of an element::
+
+        session.query(MyClass).options(
+            defaultload(MyClass.someattribute).
+            joinedload(MyOtherClass.someotherattribute)
+        )
+
+    :func:`.defaultload` is also useful for setting column-level options
+    on a related class, namely that of :func:`.defer` and :func:`.undefer`::
+
+        session.query(MyClass).options(
+            defaultload(MyClass.someattribute).
+            defer("some_column").
+            undefer("some_other_column")
+        )
+
+    .. seealso::
+
+        :ref:`relationship_loader_options`
+
+        :ref:`deferred_loading_w_multiple`
 
     """
     return loadopt.set_relationship_strategy(
@@ -903,7 +948,7 @@ def defaultload(*keys):
 
 @loader_option()
 def defer(loadopt, key):
-    """Indicate that the given column-oriented attribute should be deferred, e.g.
+    r"""Indicate that the given column-oriented attribute should be deferred, e.g.
     not loaded until accessed.
 
     This function is part of the :class:`.Load` interface and supports
@@ -967,7 +1012,7 @@ def defer(key, *addl_attrs):
 
 @loader_option()
 def undefer(loadopt, key):
-    """Indicate that the given column-oriented attribute should be undeferred,
+    r"""Indicate that the given column-oriented attribute should be undeferred,
     e.g. specified within the SELECT statement of the entity as a whole.
 
     The column being undeferred is typically set up on the mapping as a
@@ -1043,11 +1088,11 @@ def undefer_group(loadopt, name):
         :func:`.orm.undefer`
 
     """
-    loadopt._merge_into_path = True
     return loadopt.set_column_strategy(
         "*",
         None,
-        {"undefer_group_%s" % name: True}
+        {"undefer_group_%s" % name: True},
+        opts_only=True
     )
 
 
